/*
 *  dsi.c -- Development side includes
 *
 *  Copyright (c) All Rights Reserved. See details at the end of the file.
 */

/******************************* Documentation ********************************/
/*
 *  usage:  dsiPatch [-I incDir] [< fileList] [files ...]
 */
/********************************* Includes ***********************************/

#include    "posix.h"

/********************************* Defines ************************************/

#define MAX_LINE    (64 * 1024)
#define MAX_FNAME   (4096)

/********************************** Locals ************************************/

static int      finished;
static int      verbose;
static char     parentDir[MAX_FNAME];
static char     *incDir;

/**************************** Forward Declarations ****************************/

static void     openSignals();
static void     catchInterrupt(int signo);
static int      patchFileList(FILE *fp);
static int      patch(char *path);
static char     *getNextTok(char *tokBuf, int tokBufSize, char *start);
static int      matchLink(FILE *ofp, char *tag, char **start);

/************************************ Code ************************************/

int main(int argc, char *argv[])
{
    char        *argp;
    int         errors, i, nextArg;

    verbose = errors = 0;
    incDir = ".";

    for (nextArg = 1; nextArg < argc; nextArg++) {
        argp = argv[nextArg];
        if (*argp != '-') {
            break;
        }
        if (strcmp(argp, "-I") == 0) {
            if (nextArg >= argc) {
                errors++;
            } else {
                incDir = argv[++nextArg];
            }

        } else if (strcmp(argp, "-v") == 0) {
            verbose++;
        }
    }
    if (errors) {
        fprintf(stderr, "dsi: usage: [-I incDir] [< fileList] files ...\n");
        exit(2);
    }

    openSignals();

    if (nextArg >= argc) {
        if (patchFileList(stdin) < 0) {
            return -1;
        }
        
    } else {
        for (i = nextArg; !finished && i < argc; i++) {
            if (patch(argv[i]) < 0) {
                return -1;
            }
        }
    }
    return 0;
}


/*
 *  Read a list of files from stdin and patch each in-turn. Keep going on errors.
 */
static int patchFileList(FILE *fp)
{
    char    buf[MAX_FNAME];
    char    *cp;
    size_t  len;

    while (!finished && !feof(fp)) {
        if (fgets(buf, sizeof(buf), fp) == 0) {
            break;
        }
        cp = buf;
        for (cp = buf; *cp && isspace((int) *cp); ) {
            cp++;
        }
        len = strlen(cp);
        if (cp[len - 1] == '\n') {
            cp[len - 1] = '\0';
        }
        if (patch(cp) < 0) {
            return -1;
        }
    }
    return 0;
}


/*
 *  Patch a file
 */
static int patch(char *path)
{
    struct stat sbuf;
    FILE        *ifp, *ofp, *dsiFp;
    char        dsiName[MAX_FNAME], dsiPath[MAX_FNAME];
    char        tmpFile[MAX_FNAME], saveFile[MAX_FNAME];
    char        searchPat[MAX_FNAME], truePat[MAX_FNAME], falsePat[MAX_FNAME];
    char        *dsiBuf, *cp, *ep, *start, *end, *tok, *tag, *ext, *inBuf;
    char        *startDsi, *ip;
    size_t      rc, len;
    int         c, i, level, line, patched;
    
    dsiBuf = 0;
    patched = 0;
    ifp = ofp = dsiFp = 0;

    if (*path == '.' && path[1] == '/') {
        path += 2;
    }
    if (*path == '/') {
        fprintf(stderr, "dsi: Path must be relative: %s\n", path);
        return -1;
    }
    for (level = 0, cp = path; *cp; cp++) {
        if (*cp == '/') {
            level++;
        }
    }
    if (verbose) {
        printf("Patching %s\n", path);
    }
    cp = parentDir;
    for (i = 0; i < level; i++) {
        *cp++ = '.';
        *cp++ = '.';
        *cp++ = '/';
    }
    *cp = '\0';
            
    ifp = fopen(path, "r" MPR_TEXT);
    if (ifp == 0) {
        fprintf(stderr, "dsi: Can't open %s\n", path);
        return -1;
    }
    stat(path, &sbuf);
    inBuf = (char*) malloc((size_t) sbuf.st_size + 1);
    rc = fread(inBuf, 1, (size_t) sbuf.st_size, ifp);
    if (rc <= 0) {
        fprintf(stderr, "dsi: Can't read file %s\n", path);
        free(inBuf);
        return -1;
    }
    inBuf[rc] = '\0';

    sprintf(tmpFile, "%s.new", path);
    ofp = fopen(tmpFile, "w" MPR_TEXT);
    if (ofp == 0) {
        fprintf(stderr, "dsi: Can't open %s\n", tmpFile);
        goto error;
    }

    line = 0;
    for (ip = inBuf; *ip; ip++) {
        start = ip;
        if (dsiFp == 0) {
            if (*ip == '\n') {
                /* Remove duplicate new lines */
                while (ip[1] == '\n') {
                    ip++;
                }
                fputc(*ip, ofp);
                line++;
                continue;
            }
            if (*ip != '<' || ip[1] != '!' || ip[2] != '-' || ip[3] != '-') {
                fputc(*ip, ofp);
                continue;
            }

            tok = "<!-- BeginDsi \"";
            len = strlen(tok);
            if (strncmp(start, tok, strlen(tok)) == 0) {
                /*
                 *  Cleanup if file has been corrupted by HTML editors
                 */
                if (start > inBuf && start[-1] != '\n') {
                    fprintf(ofp, "\n");
                }

                startDsi = start;
                start += len - 1;

                end = getNextTok(dsiName, sizeof(dsiName), start);
                if (end == 0) {
                    fprintf(stderr, 
                        "dsi: Syntax error for DSI in %s at line %d\n", path, line);
                    goto error;
                }
                for (start = end; *start && isspace((int) *start); start++);

                /*
                 *  Parse any pattern match args
                 *      "searchPattern" "trueReplace" "falseReplace"
                 */
                if (*start == '"') {
                    start = getNextTok(searchPat, sizeof(searchPat), start);
                    start = getNextTok(truePat, sizeof(truePat), start);
                    start = getNextTok(falsePat, sizeof(falsePat), start);
                }
                if (start == 0) {
                    fprintf(stderr, "dsi: Syntax error for DSI in %s at line %d\n", path, line);
                    goto error;
                }

                for (; *start && isspace((int) *start); start++);
                end = strstr(start, "-->");
                if (end == 0) {
                    fprintf(stderr, "dsi: Missing closing comment in %s at line %d\n", path, line);
                    goto error;
                }
                c = end[3];
                end[3] = '\0';
                fprintf(ofp, "%s\n", startDsi);
                end[3] = c;

                if (dsiName[0] != '/') {
                    sprintf(dsiPath, "%s/%s", incDir, dsiName);
                } else {
                    strcpy(dsiPath, dsiName);
                }
                dsiFp = fopen(dsiPath, "r" MPR_TEXT);
                if (dsiFp == 0) {
                    fprintf(stderr, "dsi: Can't open DSI %s. Referenced in %s at line %d\n", dsiPath, path, line);
                    goto error;
                }
                stat(dsiPath, &sbuf);

                if (verbose > 1) {
                    printf("     DSI %s\n", dsiPath);
                }
                ip = &end[2];

            } else {
                fputc(*ip, ofp);
            }

        } else {
            tok = "<!-- EndDsi -->";
            len = strlen(tok);
            if (strncmp(start, tok, strlen(tok)) == 0) {
                dsiBuf = (char*) malloc((size_t) sbuf.st_size + 1);
                rc = fread(dsiBuf, 1, (size_t) sbuf.st_size, dsiFp);
                if (rc <= 0) {
                    fprintf(stderr, "dsi: Can't read DSI %s\n", dsiPath);
                    return -1;
                }
                dsiBuf[rc] = '\0';
                fclose(dsiFp);
                dsiFp = 0;              
                patched++;

                if (level == 0 && 0) {
                    rc = fwrite(dsiBuf, 1, rc, ofp);

                } else {
                    for (cp = dsiBuf; *cp; ) {
                        /*
                         *  This is a very fragile parser. It really needs to be more flexible.
                         */

                        /*
                         *  <a href=
                         *  <link .... href=
                         */
                        if (matchLink(ofp, " href=\"", &cp)) {
                            continue;
                        } 
                        if (matchLink(ofp, "\thref=\"", &cp)) {
                            continue;
                        } 

                        /*
                         *  <img src=
                         *  <script ... src=
                         */
                        if (matchLink(ofp, " src=\"", &cp)) {
                            continue;
                        } 
                        if (matchLink(ofp, "\tsrc=\"", &cp)) {
                            continue;
                        } 

                        /*
                         *  <form ... action=
                         */
                        if (matchLink(ofp, "action=\"", &cp)) {
                            continue;
                        } 

                        /*
                         *  _ROOT_ = "
                         */
                        if (matchLink(ofp, "_ROOT_=\"", &cp)) {
                            continue;
                        } 

                        /*
                         *  Special _DSI_ pattern editing
                         */
                        tag = "_DSI_";
                        len = strlen(tag);
                        if (strncmp(cp, tag, len) == 0) {
                            cp += len - 1;
                            ep = strchr(cp, '"');
                            if (strncmp(cp, searchPat, strlen(searchPat)) == 0){
                                fprintf(ofp, "%s\"", truePat);

                            } else {
                                fprintf(ofp, "%s\"", falsePat);
                            }
                            cp = ep + 1;
                            continue;
                        }

                        /*
                         *  Javascript patches
                         */
                        ext = strchr(dsiPath, '.');
                        if (ext && strcmp(ext, ".js") == 0) {
                            /*
                             *  = / *DSI* / "
                             */
                            if (matchLink(ofp, "= /*DSI*/ \"", &cp)) {
                                continue;
                            } 

                            /*
                             *  src = '
                             */
                            if (matchLink(ofp, "src == '", &cp)) {
                                continue;
                            } 
                        }

                        fputc(*cp, ofp);
                        cp++;

                    }
                }
                fprintf(ofp, "%s", tok);
                ip = &start[strlen(tok)];
                
                /*
                 *  Cleanup if file has been corrupted by HTML editors
                 */
                if (ip[1] != '\0' && ip[2] != '\0' && ip[2] != '\n') {
                    fputc('\n', ofp);
                }

            } else {
                /*  Don't output as it is being replaced with DSI content */
            }
        }
    }
    fclose(ifp);

    if (! patched) {
        fclose(ofp);
        unlink(tmpFile);

    } else {
        /*
         *  If we found a DSI in the file, rename the patched file
         */
        fclose(ofp);
        sprintf(saveFile, "%s.dsiSave", path);
        unlink(saveFile);
        chmod(path, 0755);
        rename(path, saveFile);
        if (rename(tmpFile, path) < 0) {
            fprintf(stderr, "dsi: Can't rename %s to %s\n", tmpFile, path);
            rename(saveFile, path);
            return -1;
        }
        unlink(saveFile);
    }
    return 0;

error:
    if (ifp) {
        fclose(ifp);
    }
    if (ofp) {
        fclose(ofp);
    }
    if (inBuf) {
        free(inBuf);
    }
    if (dsiBuf) {
        free(dsiBuf);
    }
    return -1;
}
    

/*
 *  Initialize signals
 */
static void openSignals() 
{
#if !_WIN32
    struct sigaction    act;

    act.sa_flags = 0;
    sigemptyset(&act.sa_mask);

    act.sa_handler = catchInterrupt;
    sigaction(SIGINT, &act, 0);
#endif
}


static void catchInterrupt(int signo)
{
    finished++;
}


/*
 *  Parse the next quoted string in the input beginning at start. Return the token (minus quotes) in the 
 *  tokBuf. Return with the input pointer one past the trailing quote. Return 0 on all errors.
 */
static char *getNextTok(char *tokBuf, int tokBufSize, char *start)
{
    char    *end;

    if (start == 0 || *start == '\0') {
        return 0;
    }

    while (*start && isspace((int) *start)) {
        start++;
    }

    if (*start != '"') {
        fprintf(stderr, "dsi: Missing opening quote\n");
        return 0;
    }

    start++;
    for (end = start; *end && *end != '"'; ) {
        end++;
    }
    if (*end != '"') {
        fprintf(stderr, "dsi: Missing closing quote\n");
        return 0;
    }
    if ((end - start) >= (tokBufSize - 1)) {
        fprintf(stderr, "dsi: Token too big\n");
        return 0;
    }
    strncpy(tokBuf, start, end - start);
    tokBuf[end - start] = '\0';
    return end + 1;
}


/*
 *  Match a link that needs patching 
 */
static int matchLink(FILE *ofp, char *tag, char **start)
{
    size_t  len;

    len = strlen(tag);
    if (strncmp(*start, tag, len) == 0) {
        *start += len;
        if ((*start)[0] == '#' || strncmp(*start, "http://", 7) == 0) {
            fprintf(ofp, "%s", tag);
        } else {
            fprintf(ofp, "%s%s", tag, parentDir);
        }
        return 1;
    }
    return 0;
}

/*
 *  @copy   default
 *  
 *  Copyright (c) Embedthis Software LLC, 2003-2012. All Rights Reserved.
 *  Copyright (c) Michael O'Brien, 1993-2012. All Rights Reserved.
 *  
 *  This software is distributed under commercial and open source licenses.
 *  You may use the GPL open source license described below or you may acquire 
 *  a commerciar license from Embedthis Software. You agree to be fully bound 
 *  by the terms of either license. Consult the LICENSE.TXT distributed with 
 *  this software for full details.
 *  
 *  This software is open source; you can redistribute it and/or modify it 
 *  under the terms of the GNU General Public License as published by the 
 *  Free Software Foundation; either version 2 of the License, or (at your 
 *  option) any later version. See the GNU General Public License for more 
 *  details at: http://www.embedthis.com/downloads/gplLicense.html
 *  
 *  This program is distributed WITHOUT ANY WARRANTY; without even the 
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 *  
 *  This GPL license does NOT permit incorporating this software into 
 *  proprietary programs. If you are unable to comply with the GPL, you must
 *  acquire a commercial license to use this software. Commercial licenses 
 *  for this software and support services are available from Embedthis 
 *  Software at http://www.embedthis.com 
 *
 *  Local variables:
    tab-width: 4
    c-basic-offset: 4
    End:
    vim: sw=4 ts=4 expandtab

    @end
 */
